import logging
import sys
import importlib
import subprocess
import requests
from lxml import html

from kestrel.exceptions import DataSourceError


_logger = logging.getLogger(__name__)


XPATH_PYPI_PKG_HOME = "/html/body/main/div[4]/div/div/div[1]/div[2]/ul/li[1]/a/@href"
XPATH_PYPI_PKG_SOURCE = "/html/body/main/div[4]/div/div/div[1]/div[2]/ul/li[2]/a/@href"
STIX_SHIFTER_HOMEPAGE = "https://github.com/opencybersecurityalliance/stix-shifter"


def get_package_name(connector_name):
    return "stix-shifter-modules-" + connector_name.replace("_", "-")


def verify_package_origin(connector_name):
    _logger.debug("go to PyPI to verify package genuineness from STIX-shifter project")
    package_name = get_package_name(connector_name)

    try:
        pypi_response = requests.get(f"https://pypi.org/project/{package_name}")
        pypi_etree = html.fromstring(pypi_response.content)
    except:
        raise DataSourceError(
            f'STIX-shifter connector for "{connector_name}" is not installed '
            f'and Kestrel guessed Python package name "{package_name}" but failed to locate it at PyPI',
            "please manually install the correct STIX-shifter connector Python package.",
        )

    try:
        p_homepage = pypi_etree.xpath(XPATH_PYPI_PKG_HOME)[0]
        p_source = pypi_etree.xpath(XPATH_PYPI_PKG_SOURCE)[0]
    except:
        raise DataSourceError(
            f'STIX-shifter connector for "{connector_name}" is not installed '
            f'and Kestrel guessed Python package name "{package_name}" but could not verify its genuineness due to PyPI design change',
            "please find the correct STIX-shifter connector Python package to install. "
            "And report to Kestrel developers about this package verification failure",
        )

    if p_homepage != STIX_SHIFTER_HOMEPAGE or p_source != STIX_SHIFTER_HOMEPAGE:
        raise DataSourceError(
            f'STIX-shifter connector for "{connector_name}" is not installed '
            f'and Kestrel found Python package "{package_name}" is not a genuine STIX-shifter package',
            "please find the correct STIX-shifter connector Python package to install. "
            "And report to Kestrel developers about this malicious package",
        )

    _logger.info(f'"{package_name}" verified as a STIX-shifter package.')


def check_module_availability(connector_name):
    try:
        importlib.import_module(
            "stix_shifter_modules." + connector_name + ".entry_point"
        )
    except:
        _logger.info(f'miss STIX-shifter connector "{connector_name}"')

        package_name = get_package_name(connector_name)
        _logger.debug(f"guess the connector package name: {package_name}")

        verify_package_origin(connector_name)

        _logger.info(f'install Python package "{package_name}".')
        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", package_name]
            )
        except:
            _logger.info("package installation with 'pip' failed.")

        try:
            importlib.import_module(
                "stix_shifter_modules." + connector_name + ".entry_point"
            )
        except:
            raise DataSourceError(
                f'STIX-shifter connector for "{connector_name}" is not installed '
                f'and Kestrel failed to install the possible Python package "{package_name}"',
                "please manually install the corresponding STIX-shifter connector Python package.",
            )



"""A Kestrel session provides an isolated stateful runtime space for a huntflow.

A huntflow is the source code or script of a cyber threat hunt, which can be
developed offline in a text editor or interactively as the hunt goes. A Kestrel
session provides the runtime space for a huntflow that allows execution and
inspection of hunt statements in the huntflow. The :class:`Session` class in this
module supports both non-interactive and interactive execution of huntflows as
well as comprehensive APIs besides execution.

.. highlight:: python

Examples:
    A non-interactive execution of a huntflow::

        from kestrel.session import Session
        with Session() as session:
            open(huntflow_file) as hff:
                huntflow = hff.read()
            session.execute(huntflow)

    An interactive composition and execution of a huntflow::

        from kestrel.session import Session
        with Session() as session:
            try:
                hunt_statement = input(">>> ")
            except EOFError:
                print()
                break
            else:
                output = session.execute(hunt_statement)
                print(output)

    Export Kestrel variable to Python::

        from kestrel.session import Session
        huntflow = ""\"newvar = GET process
                               FROM stixshifter://workstationX
                               WHERE [process:name = 'cmd.exe']""\"
        with Session() as session:
            session.execute(huntflow)
            cmds = session.get_variable("newvar")
        for process in cmds:
            print(process["name"])

"""

import tempfile
import os
import pathlib
import shutil
import uuid
import logging
import re
import time
import math
import lark
import atexit
from datetime import datetime
from contextlib import AbstractContextManager

from kestrel.exceptions import (
    KestrelSyntaxError,
    InvalidStixPattern,
)
from kestrel.syntax.parser import get_all_input_var_names
from kestrel.syntax.parser import parse
from kestrel.syntax.utils import (
    get_entity_types,
    get_keywords,
    all_relations,
    LITERALS,
    AGG_FUNCS,
    TRANSFORMS,
)
from kestrel.semantics import *
from kestrel.codegen import commands
from kestrel.codegen.display import DisplayBlockSummary
from kestrel.codegen.summary import gen_variable_summary
from firepit import get_storage
from firepit.exceptions import StixPatternError
from kestrel.utils import set_current_working_directory
from kestrel.config import load_config
from kestrel.datasource import DataSourceManager
from kestrel.analytics import AnalyticsManager

_logger = logging.getLogger(__name__)


class Session(AbstractContextManager):
    """Kestrel Session class

    A session object needs to be instantiated to create a Kestrel runtime space.
    This is the foundation of multi-user dynamic composition and execution of
    huntflows. A Kestrel session has two important properties:

    - Stateful: a session keeps track of states/effects of statements that have
      been previously executed in this session, e.g., the values of previous
      established Kestrel variables. A session can invoke more than one
      :meth:`execute`, and each :meth:`execute` can process a block of Kestrel code,
      i.e., multiple Kestrel statements.

    - Isolated: each session is established in an isolated space (memory and
      file system):

      - Memory isolation is accomplished by OS process and memory space
        management automatically -- different Kestrel session instances will not
        overlap in memory.

      - File system isolation is accomplished with the setup and management of
        a temporary runtime directory for each session.

    Args:

      runtime_dir (str): to be used for :attr:`runtime_directory`.

      store_path (str): the file path or URL to initialize :attr:`store`.

      debug_mode (bool): to be assign to :attr:`debug_mode`.

    Attributes:

        session_id (str): The Kestrel session ID, which will be created as a random
          UUID if not given in the constructor.

        runtime_directory (str): The runtime directory stores session related
          data in the file system such as local cache of queried results,
          session log, and may be the internal store. The session will use
          a temporary directory derived from :attr:`session_id` if the path is
          not specified in constructor parameters.

        store (firepit.SqlStorage): The internal store used
          by the session to normalize queried results, implement cache, and
          realize the low level code generation. The store from the
          ``firepit`` package provides an operation abstraction
          over the raw internal database: either a local store, e.g., SQLite,
          or a remote one, e.g., PostgreSQL. If not specified from the
          constructor parameter, the session will use the default SQLite
          store in the :attr:`runtime_directory`.

        debug_mode (bool): The debug flag set by the session constructor. If
          True, a fixed debug link ``/tmp/kestrel`` of :attr:`runtime_directory`
          will be created, and :attr:`runtime_directory` will not be removed by
          the session when terminating.

        runtime_directory_is_owned_by_upper_layer (bool): The flag to specify
          who owns and manages :attr:`runtime_directory`. False by default,
          where the Kestrel session will manage session file system isolation --
          create and destory :attr:`runtime_directory`. If True, the runtime
          directory is created, passed in to the session constructor, and will
          be destroyed by the calling site.

        symtable (dict): The continuously updated *symbol table* of the running
          session, which is a dictionary mapping from Kestrel variable names
          ``str`` to their associated Kestrel internal data structure
          ``VarStruct``.

        data_source_manager (kestrel.datasource.DataSourceManager): The
          data source manager handles queries to all data source interfaces such as
          local file stix bundle and stix-shifter. It also stores previous
          queried data sources for the session, which is used for a syntax
          sugar when there is no data source in a Kestrel ``GET`` statement -- the
          last data source is implicitly used.

        analytics_manager (kestrel.analytics.AnalyticsManager): The analytics
          manager handles all analytics related operations such as executing an
          analytics or getting the list of analytics for code auto-completion.

    """

    def __init__(
        self, session_id=None, runtime_dir=None, store_path=None, debug_mode=False
    ):
        _logger.debug(
            f"Establish session with session_id: {session_id}, runtime_dir: {runtime_dir}, store_path:{store_path}, debug_mode:{debug_mode}"
        )

        self.config = load_config()

        if session_id:
            self.session_id = session_id
        else:
            self.session_id = str(uuid.uuid4())

        self.debug_mode = (
            True
            if debug_mode or os.getenv(self.config["debug"]["env_var"], False)
            else False
        )

        # default value of runtime_directory ownership
        self.runtime_directory_is_owned_by_upper_layer = False

        # runtime (temporary) directory to store session-related data
        sys_tmp_dir = pathlib.Path(tempfile.gettempdir())
        if runtime_dir:
            if os.path.exists(runtime_dir):
                self.runtime_directory_is_owned_by_upper_layer = True
            else:
                pathlib.Path(runtime_dir).mkdir(parents=True, exist_ok=True)
            self.runtime_directory = runtime_dir
        else:
            tmp_dir = sys_tmp_dir / (
                self.config["session"]["cache_directory_prefix"] + self.session_id
            )
            self.runtime_directory = tmp_dir.expanduser().resolve()
            if tmp_dir.exists():
                if tmp_dir.is_dir():
                    _logger.debug(
                        "Kestrel session with runtime_directory exists, reuse it."
                    )
                else:
                    _logger.debug(
                        "strange tmp file that uses kestrel session dir name, remove it."
                    )
                    os.remove(self.runtime_directory)
            else:
                _logger.debug(
                    f"create new session runtime_directory: {self.runtime_directory}."
                )
                tmp_dir.mkdir(parents=True, exist_ok=True)

        if self.debug_mode:
            runtime_directory_master = (
                sys_tmp_dir / self.config["debug"]["cache_directory"]
            )

            # runtime_directory_master.exists() should not be used
            # it will return False for broken link
            try:
                runtime_directory_master.unlink()
            except FileNotFoundError:
                pass

            runtime_directory_master.symlink_to(self.runtime_directory)

        # local database of SQLite or PostgreSQL
        if not store_path:
            # use the default local database in config.py
            local_database_path = self.config["session"]["local_database_path"]
            if "://" in local_database_path:
                store_path = local_database_path
            else:
                store_path = os.path.join(self.runtime_directory, local_database_path)
        self.store = get_storage(store_path, self.session_id)

        # Symbol Table
        # linking variables in syntax with internal data structure
        # handling fallback_var for the most recently accessed var
        # {"var": VarStruct}
        self.symtable = {}

        self.data_source_manager = DataSourceManager(self.config)
        self.analytics_manager = AnalyticsManager(self.config)
        iso_ts_regex = r"\d{4}(-\d{2}(-\d{2}(T\d{2}(:\d{2}(:\d{2}Z?)?)?)?)?)?"
        self._iso_ts = re.compile(iso_ts_regex)

        atexit.register(self.close)

    def execute(self, codeblock):
        """Execute a Kestrel code block.

        A Kestrel statement or multiple consecutive statements constitute a code
        block, which can be executed by this method. New Kestrel variables can be
        created in a code block such as ``newvar = GET ...``. Two types of Kestrel
        variables can be legally referred in a Kestrel statement in the code block:

        * A Kestrel variable created in the same code block prior to the reference.

        * A Kestrel variable created in code blocks previously executed by the
          session. The session maintains the :attr:`symtable` to keep the state
          of all previously executed Kestrel statements and their established Kestrel
          variables.

        Args:
            codeblock (str): the code block to be executed.

        Returns:
            A list of outputs that each of them is the output for each
            statement in the inputted code block.
        """
        ast = self.parse(codeblock)
        return self._execute_ast(ast)

    def parse(self, codeblock):
        """Parse a Kestrel code block.

        Parse one or multiple consecutive Kestrel statements (a Kestrel code block)
        into the abstract syntax tree. This could be useful for frontends that
        need to parse a statement *without* executing it in order to render
        some type of interface.

        Args:
            codeblock (str): the code block to be parsed.

        Returns:
            A list of dictionaries that each of them is an *abstract syntax
            tree* for one Kestrel statement in the inputted code block.
        """
        try:
            ast = parse(
                codeblock,
                self.config["language"]["default_variable"],
                self.config["language"]["default_sort_order"],
            )
        except lark.UnexpectedEOF as err:
            raise KestrelSyntaxError(
                err.line, err.column, "end of line", "", err.expected
            )
        except lark.UnexpectedCharacters as err:
            raise KestrelSyntaxError(
                err.line, err.column, "character", err.char, err.allowed
            )
        except lark.UnexpectedToken as err:
            raise KestrelSyntaxError(
                err.line, err.column, "token", err.token, err.accepts or err.expected
            )
        return ast

    def get_variable_names(self):
        """Get the list of Kestrel variable names created in this session."""
        return list(self.symtable.keys())

    def get_variable(self, var_name):
        """Get the data of Kestrel variable ``var_name``, which is list of homogeneous entities (STIX SCOs)."""
        # In the future, consider returning a generator here?
        return self.symtable[var_name].get_entities()

    def create_variable(self, var_name, objects, object_type=None):
        """Create a new Kestrel variable ``var_name`` with data in ``objects``.

        This is the API equivalent to Kestrel command ``NEW``, while allowing more
        flexible objects types (Python objects) than the objects serialized
        into text/JSON in the command ``NEW``.

        Args:

            var_name (str): The Kestrel variable to be created.

            objects (list): List of Python objects, currently support either a
              list of ``str`` or a list of ``dict``.

            object_type (str): The Kestrel entity type for the created Kestrel
              variable. It overrides the ``type`` field in ``objects``. If
              there is no ``type`` field in ``objects``, e.g., ``objects`` is a
              list of ``str``, this parameter is required.

        """
        virtual_stmt_ast = [
            {"command": "new", "output": var_name, "data": objects, "type": object_type}
        ]
        self._execute_ast(virtual_stmt_ast)

    def do_complete(self, code, cursor_pos):
        """Kestrel code auto-completion.

        This function gives a list of suggestions on the inputted partial Kestrel
        code to complete it. The current version sets the context for
        completion on word level -- it will reason around the last word in the
        input Kestrel code to provide suggestions. Data sources and analytics names
        can also be completed since the entire URI are single words (no space
        in data source or analytic name string). This feature can be used to
        list all available data sources or analytics, e.g., giving the last
        partial word ``stixshifter://``.

        Currently this method computes code completion based on:

        * Kestrel keywords
        * Kestrel variables
        * data source names
        * analytics names

        Args:
            code (str): Kestrel code.
            cursor_pos (int): the position to start completion (index in ``code``).

        Returns:
            A list of suggested strings to complete the code.
        """
        prefix = code[:cursor_pos]
        words = prefix.split(" ")
        last_word = words[-1]
        last_char = prefix[-1]
        _logger.debug('code="%s" prefix="%s" last_word="%s"', code, prefix, last_word)

        if "START" in prefix or "STOP" in prefix:
            return self._get_complete_timestamp(last_word)
        elif "://" in last_word:
            scheme, path = last_word.split("://")
            if scheme in self.data_source_manager.schemes():
                data_source_names = (
                    self.data_source_manager.list_data_sources_from_scheme(scheme)
                )
                allnames = [scheme + "://" + name for name in data_source_names]
                _logger.debug(
                    f"auto-complete from data source interface {scheme}: {allnames}"
                )
            elif scheme in self.analytics_manager.schemes():
                analytics_names = self.analytics_manager.list_analytics_from_scheme(
                    scheme
                )
                allnames = [scheme + "://" + name for name in analytics_names]
                _logger.debug(
                    f"auto-complete from analytics interface {scheme}: {allnames}"
                )
            else:
                allnames = []
                _logger.debug("cannot find auto-complete interface")
        else:
            _logger.debug("standard auto-complete")

            try:
                stmt = self.parse(prefix)
                _logger.debug("first parse: %s", stmt)
                last_stmt = stmt[-1]
                if last_stmt["command"] == "assign" and last_stmt["output"] == "_":
                    # Special case for a varname alone on a line
                    allnames = [
                        v for v in self.get_variable_names() if v.startswith(prefix)
                    ]
                    if not allnames:
                        return ["=", "+"] if prefix.endswith(" ") else []

                # If it parses successfully, add something so it will fail
                self.parse(prefix + " @autocompletions@")
            except KestrelSyntaxError as e:
                _logger.debug("exception: %s", e)
                varnames = self.get_variable_names()
                keywords = set(get_keywords())
                _logger.debug("keywords: %s", keywords)
                tmp = []
                for token in e.expected:
                    _logger.debug("token: %s", token)
                    if token == "VARIABLE":
                        tmp.extend(varnames)
                    elif token == "DATASRC":
                        schemes = self.data_source_manager.schemes()
                        tmp.extend([f"{scheme}://" for scheme in schemes])
                        tmp.extend(varnames)
                    elif token == "ANALYTICS":
                        schemes = self.analytics_manager.schemes()
                        tmp.extend([f"{scheme}://" for scheme in schemes])
                    elif token == "ENTITY_TYPE":
                        tmp.extend(get_entity_types())
                    elif token.startswith("STIXPATH"):
                        # TODO: figure out the varname and get its attrs
                        continue
                    elif token.startswith("STIXPATTERNBODY"):
                        # TODO: figure out how to complete STIX patterns
                        continue
                    elif token == "RELATION":
                        if last_word:
                            tmp.extend(get_entity_types())
                        else:
                            tmp.extend(all_relations)
                    elif token == "BY":
                        tmp.append("BY")
                    elif token == "REVERSED":
                        if last_char == " ":
                            tmp.append("BY")
                        else:
                            # "procs = FIND process l" will expect ['REVERSED', 'VARIABLE']
                            # override results from the case of VARIABLE
                            tmp = all_relations
                            break
                    elif token == "FUNCNAME":
                        tmp.extend(AGG_FUNCS)
                    elif token == "TRANSFORM":
                        tmp.extend(TRANSFORMS)
                    elif token in LITERALS:
                        continue
                    elif token.startswith("__ANON"):
                        continue
                    elif token == "EQUAL":
                        tmp.append("=")
                    elif token in keywords and last_word.islower():
                        # keywords has both upper and lower case
                        tmp.append(token.lower())
                    else:
                        tmp.append(token)
                allnames = sorted(tmp)

        suggestions = [
            name[len(last_word) :] for name in allnames if name.startswith(last_word)
        ]
        _logger.debug("%s -> %s", allnames, suggestions)
        return suggestions

    def close(self):
        """Explicitly close the session.

        This may be executed by a context manager or when the program exits.
        """
        # this subroutine could be invoked twice by a context manager and program exit.
        # only execute it once (when self.store not deleted).
        if hasattr(self, "store"):

            # release resources
            self.store.close()
            del self.store

            # manage temp folder for debug
            if not self.runtime_directory_is_owned_by_upper_layer:
                if self.debug_mode:
                    self._leave_exit_marker()
                    self._remove_obsolete_debug_folders()
                else:
                    shutil.rmtree(self.runtime_directory)

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def _execute_ast(self, ast):
        displays = []
        new_vars = []

        start_exec_ts = time.time()
        for stmt in ast:

            try:

                # pre-processing: semantics check and completion
                #   - ensure all parsed elements not empty
                #   - check existance of argument variables
                #   - complete data source if omitted by user
                #   - complete input context
                check_elements_not_empty(stmt)
                for input_var_name in get_all_input_var_names(stmt):
                    check_var_exists(input_var_name, self.symtable)
                if stmt["command"] == "get":
                    recognize_var_source(stmt, self.symtable)
                    complete_data_source(
                        stmt, self.data_source_manager.queried_data_sources[-1]
                    )
                if stmt["command"] == "load" or stmt["command"] == "save":
                    stmt["path"] = pathlib.Path(stmt["path"]).expanduser().resolve()
                if stmt["command"] == "find":
                    check_semantics_on_find(stmt, self.symtable[stmt["input"]].type)
                if "attrs" in stmt:
                    var_struct = self.symtable[stmt["input"]]
                    stmt["attrs"] = normalize_attrs(stmt, var_struct)

                # code generation and execution
                execute_cmd = getattr(commands, stmt["command"])

                # set current working directory for each command execution
                # use this to implicitly pass runtime_dir as an argument to each command
                # the context manager switch back cwd when the command execution completes
                with set_current_working_directory(self.runtime_directory):
                    output_var_struct, display = execute_cmd(stmt, self)

            # exception completion
            except StixPatternError as e:
                raise InvalidStixPattern(e.stix) from e

            # post-processing: symbol table update
            if output_var_struct is not None:
                output_var_name = stmt["output"]
                self._update_symbol_table(output_var_name, output_var_struct)

                if output_var_name != self.config["language"]["default_variable"]:
                    if output_var_name in new_vars:
                        new_vars.remove(output_var_name)
                    new_vars.append(output_var_name)

            if display is not None:
                displays.append(display)

        end_exec_ts = time.time()
        execution_time_sec = math.ceil(end_exec_ts - start_exec_ts)

        if self.config["session"]["show_execution_summary"] and new_vars:
            vars_summary = [
                gen_variable_summary(vname, self.symtable[vname]) for vname in new_vars
            ]
            displays.append(DisplayBlockSummary(vars_summary, execution_time_sec))

        return displays

    def _update_symbol_table(self, output_var_name, output_var_struct):
        self.symtable[output_var_name] = output_var_struct
        self.symtable[self.config["language"]["default_variable"]] = output_var_struct

    def _leave_exit_marker(self):
        exit_marker = os.path.join(
            self.runtime_directory, self.config["debug"]["session_exit_marker"]
        )
        with open(exit_marker, "w"):
            pass

    def _remove_obsolete_debug_folders(self):
        # will only clean debug cache directories under system temp directory

        # [(cache_dir, timestamp)]
        exited_sessions = []

        for x in pathlib.Path(tempfile.gettempdir()).iterdir():
            if x.is_dir() and x.parts[-1].startswith(
                self.config["session"]["cache_directory_prefix"]
            ):
                marker = x / self.config["debug"]["session_exit_marker"]
                if marker.exists():
                    exited_sessions.append((x, marker.stat().st_mtime))

        # preserve the newest self.config["debug"]["maximum_exited_session"] debug sessions
        exited_sessions.sort(key=lambda x: x[1])
        for x, _ in exited_sessions[: -self.config["debug"]["maximum_exited_session"]]:
            shutil.rmtree(x)

    def _get_complete_timestamp(self, ts_str):
        valid_ts_formats = [
            "%Y",
            "%Y-%m",
            "%Y-%m-%d",
            "%Y-%m-%dT%H",
            "%Y-%m-%dT%H:%M",
            "%Y-%m-%dT%H:%M:%S",
        ]
        complete_ts = []
        for vts in valid_ts_formats:
            ts = ts_str.split("'")[-1]
            matched = self._iso_ts.match(ts)
            if matched:
                try:
                    ts_iso = datetime.strptime(matched.group(), vts).isoformat()
                    complete_ts.append(ts_iso[len(ts) :] + "Z'")
                    if complete_ts:
                        return complete_ts
                except:
                    _logger.debug(f"Try to match timestamp {ts} by format {vts}")
                    pass



import os
import json
import logging

from kestrel.config import (
    CONFIG_DIR_DEFAULT,
    load_user_config,
)
from kestrel.utils import update_nested_dict
from kestrel.exceptions import InvalidDataSource

PROFILE_PATH_DEFAULT = CONFIG_DIR_DEFAULT / "stixshifter.yaml"
PROFILE_PATH_ENV_VAR = "KESTREL_STIXSHIFTER_CONFIG"
STIXSHIFTER_DEBUG_ENV_VAR = "KESTREL_STIXSHIFTER_DEBUG"  # debug mode for stix-shifter if the environment variable exists
ENV_VAR_PREFIX = "STIXSHIFTER_"
RETRIEVAL_BATCH_SIZE = 512

_logger = logging.getLogger(__name__)


def set_stixshifter_logging_level():
    debug_mode = os.getenv(STIXSHIFTER_DEBUG_ENV_VAR, False)
    logging_level = logging.DEBUG if debug_mode else logging.INFO
    _logger.debug(f"set stix-shifter logging level: {logging_level}")
    logging.getLogger("stix_shifter").setLevel(logging_level)
    logging.getLogger("stix_shifter_utils").setLevel(logging_level)
    logging.getLogger("stix_shifter_modules").setLevel(logging_level)


def load_profiles_from_env_var():
    env_vars = os.environ.keys()
    stixshifter_vars = filter(lambda x: x.startswith(ENV_VAR_PREFIX), env_vars)
    profiles = {}
    for evar in stixshifter_vars:
        items = evar.lower().split("_")
        suffix = items[-1]
        profile = "_".join(items[1:-1])
        _logger.debug(f"processing stix-shifter env var: {evar}:")
        if profile not in profiles:
            profiles[profile] = {}

        # decoding JSON or string values from environment variables
        if suffix == "connection" or suffix == "config":
            try:
                value = json.loads(os.environ[evar])
            except json.decoder.JSONDecodeError:
                raise InvalidDataSource(
                    profile,
                    "stixshifter",
                    f"invalid JSON in {evar} environment variable",
                )
        else:
            value = os.environ[evar]

        _logger.debug(f"profile: {profile}, suffix: {suffix}, value: {value}")
        profiles[profile][suffix] = value

    return profiles


def get_datasource_from_profiles(profile_name, profiles):
    """Validate and retrieve profile data

    Validate and retrieve profile data. The data should be a dict with
    "connector", "connection", "config" keys, and appropriate values.

    Args:
        profile_name (str): The name of the profile.
        profiles (dict): name to profile (dict) mapping.

    Returns:
        STIX-shifter config triplet
    """
    profile_name = profile_name.lower()
    if profile_name not in profiles:
        raise InvalidDataSource(
            profile_name,
            "stixshifter",
            f"no {profile_name} configuration found",
        )
    else:
        profile = profiles[profile_name]
        if not profile:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"the profile is empty",
            )
        _logger.debug(f"profile to use: {profile}")
        if "connector" not in profile:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {profile_name} connector defined",
            )
        else:
            connector_name = profile["connector"]
        if "connection" not in profile:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {profile_name} connection defined",
            )
        else:
            connection = profile["connection"]
        if "config" not in profile:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {profile_name} configuration defined",
            )
        else:
            configuration = profile["config"]
        if "host" not in connection:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {profile_name} connection section: no "host" field',
            )

        if "port" not in connection and connector_name != "stix_bundle":
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {profile_name} connection section: no "port" field',
            )

        if "auth" not in configuration:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {profile_name} configuration section: no "auth" field',
            )
    return connector_name, connection, configuration


def load_profiles():
    config = load_user_config(PROFILE_PATH_ENV_VAR, PROFILE_PATH_DEFAULT)
    if config and "profiles" in config:
        _logger.debug(f"stix-shifter profiles found in config file")
        profiles_from_file = {k.lower(): v for k, v in config["profiles"].items()}
    else:
        _logger.debug(
            "either config file does not exist or no stix-shifter profile found in config file. This may indicate a config syntax error if config file exists."
        )
        profiles_from_file = {}
    profiles_from_env_var = load_profiles_from_env_var()
    profiles = update_nested_dict(profiles_from_file, profiles_from_env_var)
    _logger.debug(f"profiles loaded: {profiles}")
    return profiles



from abc import ABC, abstractmethod
from pandas import DataFrame
from io import StringIO
import json

from kestrel.exceptions import KestrelInternalError


class AbstractDisplay(ABC):
    @abstractmethod
    def to_string(self):
        pass

    @abstractmethod
    def to_html(self):
        pass

    @abstractmethod
    def to_json(self):
        pass

    @abstractmethod
    def to_dict(self):
        pass


class DisplayDataframe(AbstractDisplay):
    def __init__(self, data):
        if isinstance(data, DataFrame):
            self.dataframe = data
        else:
            try:
                self.dataframe = DataFrame(data)
            except ValueError:
                raise KestrelInternalError("invalid input to DisplayDataframe object")

    def to_string(self):
        return self.dataframe.to_string(index=False, na_rep="")

    def to_html(self):
        escaped_df = self.dataframe.replace("\$", "\\\$", inplace=False, regex=True)
        return escaped_df.to_html(index=False, na_rep="")

    def to_json(self):
        body = self.dataframe.to_json(orient="records")
        msg = {"display": "dataframe", "data": "<<<BODY>>>"}
        return json.dumps(msg).replace('"<<<BODY>>>"', body)

    def to_dict(self):
        body = self.dataframe.fillna(0).to_dict(orient="records")
        msg = {"display": "dataframe", "data": body}
        return msg


class DisplayBlockSummary(DisplayDataframe):
    def __init__(self, vars_summary, exec_time_sec):
        self.vars_summary = vars_summary
        self.footnotes = []
        self.exec_time_sec = exec_time_sec
        summaries = []
        for summary, footnote in vars_summary:
            summaries.append(summary)
            if footnote and footnote not in self.footnotes:
                self.footnotes.append(footnote)
        super().__init__(summaries)
        self.exec_time_str = self._cal_exec_time(exec_time_sec)

    def to_string(self):
        header = f"[SUMMARY] block executed in {self.exec_time_str}"
        body = super().to_string()
        footer = "\n".join(self.footnotes)
        return "\n".join([header, body, footer])

    def to_html(self):
        header = f"<h4>Block Executed in {self.exec_time_str}</h4>"
        body = self.dataframe.to_html(index=False)
        footer = "<p>" + "<br>".join(self.footnotes) + "</p>"
        return "<div>" + header + body + footer + "</div>"

    def to_json(self):
        data = self.dataframe.to_json(orient="records")
        msg = {
            "display": "execution summary",
            "data": {
                "execution time": self.exec_time_sec,
                "variables updated": "<<<DATA>>>",
                "footnotes": self.footnotes,
            },
        }
        return json.dumps(msg).replace('"<<<DATA>>>"', data)

    def to_dict(self):
        data = self.dataframe.fillna(0).to_dict(orient="records")
        msg = {
            "display": "execution summary",
            "data": {
                "execution time": self.exec_time_sec,
                "variables updated": data,
                "footnotes": self.footnotes,
            },
        }
        return msg

    def _cal_exec_time(self, exec_time_sec):
        hours = exec_time_sec // 3600
        mins = (exec_time_sec // 60) % 60
        secs = exec_time_sec % 60
        x = f"{hours} hours, " if hours else ""
        x = f"{x}{mins} minutes, " if x or mins else ""
        x = f"{x}{secs} seconds" if x or secs else ""
        return x


class DisplayDict(AbstractDisplay):
    def __init__(self, d):
        self.dict = d
        self.dataframe = DataFrame(list(d.items()))

    def to_string(self):
        return self.dataframe.to_string(index=False, header=False, na_rep="")

    def to_html(self):
        clen = self.dataframe[0].str.len().max()
        return self.dataframe.to_html(
            index=False, header=False, na_rep="", col_space=[f"{clen}em", "1em"]
        )

    def to_json(self):
        msg = {"display": "dict", "data": self.dict}
        return json.dumps(msg)

    def to_dict(self):
        msg = {"display": "dict", "data": self.dict}
        return msg


class DisplayHtml(AbstractDisplay):
    def __init__(self, html):
        self.html = html

    def to_string(self):
        return self.html

    def to_html(self):
        return self.html

    def to_json(self):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError


class DisplayFigure(DisplayHtml):
    def __init__(self, figure):
        figure_type = type(figure)
        if (
            figure_type.__module__ + "." + figure_type.__name__
            != "matplotlib.figure.Figure"
        ):
            raise KestrelInternalError("invalid input to DisplayFigure object")
        vfile = StringIO()
        figure.savefig(vfile, format="svg")
        svg = vfile.getvalue()
        super().__init__(svg)


class DisplayWarning(AbstractDisplay):
    def __init__(self, text):
        self.text = text

    def to_string(self):
        return self.text

    def to_html(self):
        return f'<div class="warning">[WARNING] {self.text}</div>'

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_dict(self):
        return {"display": "warning", "data": self.text}



import importlib
import pkgutil
import logging
import inspect
import sys

_logger = logging.getLogger(__name__)


class InterfaceManager:
    def __init__(
        self,
        config,
        config_root_key,
        default_schema_config_path,
        module_name_prefix,
        interface_class,
        nonexist_interface_exception,
        invalid_interface_exception,
        conflict_interface_exception,
    ):
        self.scheme_to_interface = {}
        self.scheme_to_interface_name = {}
        config[config_root_key] = {}
        self.config = config
        self.config_root_key = config_root_key
        self.default_schema_config_path = default_schema_config_path
        self.nonexist_interface_exception = nonexist_interface_exception

        for name, itf in _load_interfaces(
            module_name_prefix,
            interface_class,
            invalid_interface_exception,
            conflict_interface_exception,
        ).items():
            self.scheme_to_interface.update({s: itf for s in itf.schemes()})
            self.scheme_to_interface_name.update({s: name for s in itf.schemes()})
            self.config[self.config_root_key][name] = {}

    def schemes(self):
        return list(self.scheme_to_interface.keys())

    def _parse_and_complete_uri(self, uri):
        default_schema = self._get_default_schema()
        scheme, splitter, path = uri.rpartition("://")
        if not scheme:
            scheme = default_schema
            if not splitter:
                uri = default_schema + "://" + uri
            else:
                uri = default_schema + uri
        return scheme, uri

    def _get_interface_with_config(self, scheme):
        scheme = scheme.lower()
        if scheme not in self.scheme_to_interface:
            raise self.nonexist_interface_exception(scheme)
        if scheme not in self.scheme_to_interface_name:
            raise self.nonexist_interface_exception(scheme)
        interface_name = self.scheme_to_interface_name[scheme]
        interface_config = self.config[self.config_root_key][interface_name]
        interface = self.scheme_to_interface[scheme]
        return interface, interface_config

    def _get_default_schema(self):
        # this method is required to handle dynamic config changes
        # where the default schema is changed on the fly
        partial_config_path = self.config
        for path_component in self.default_schema_config_path:
            partial_config_path = partial_config_path[path_component]
        default_schema = partial_config_path

        if default_schema not in self.scheme_to_interface:
            _logger.error(f"default datasource schema {default_schema} not found.")
            raise self.nonexist_interface_exception(default_schema)

        return default_schema


def _load_interfaces(
    module_name_prefix,
    interface_class,
    invalid_interface_exception,
    conflict_interface_exception,
):
    is_interface = _is_class(interface_class)
    interface_names = _list_interfaces(module_name_prefix)
    interfaces = {}
    for interface_name in interface_names:
        importlib.import_module(interface_name)
        cls = inspect.getmembers(sys.modules[interface_name], is_interface)
        if not cls:
            raise invalid_interface_exception(
                f'no interface class found in "{interface_name}"'
            )
        elif len(cls) > 1:
            raise invalid_interface_exception(
                f'more than one interface class found in "{interface_name}"'
            )
        else:
            interface = cls[0][1]
            interface_conflict, scheme_conflict = _search_scheme_conflict(
                interface, interfaces.values()
            )
            if interface_conflict:
                raise conflict_interface_exception(
                    interface, interface_conflict, scheme_conflict
                )
            interfaces[interface_name] = interface
    return interfaces


def _list_interfaces(module_name_prefix):
    pkg_names = map(lambda x: x.name, pkgutil.iter_modules())
    itf_names = filter(lambda x: x.startswith(module_name_prefix), pkg_names)
    return list(itf_names)


def _is_class(cls):
    return lambda obj: inspect.isclass(obj) and obj.__bases__[0] == cls


def _search_scheme_conflict(new_interface, interfaces):
    for interface in interfaces:
        for scheme_new in new_interface.schemes():
            for scheme_old in interface.schemes():
                if scheme_new == scheme_old:
                    return interface, scheme_new
    return None, None



"""The STIX bundle data source package provides access to canned data in STIX
bundles locally or remotely.

"""

import json
import logging
import re
import uuid
import pathlib
from datetime import datetime, timezone
from dateutil import parser
import requests

from stix2matcher.matcher import Pattern

from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import DataSourceManagerInternalError, DataSourceConnectionError

_logger = logging.getLogger(__name__)


def _make_query_id(uri, pattern):
    return str(uuid.uuid5(uuid.NAMESPACE_URL, str(uri) + pattern))


def _make_query_dir(query_id):
    path = pathlib.Path(query_id)
    path.mkdir(parents=True, exist_ok=False)
    return path


def _make_download_dir():
    path = pathlib.Path("downloads")
    path.mkdir(parents=True, exist_ok=True)
    return path


def fixup_pattern(pattern):
    # The matcher doesn't accept TimestampLiterals in START/STOP
    # See https://github.com/oasis-open/cti-pattern-validator/issues/52
    return re.sub(r"(START|STOP)\s+t'", r"\1 '", pattern)


class StixBundleInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX bundle data source interface supporting ``file:///``, ``http://``, ``https://`` scheme."""
        return ["file", "http", "https"]

    @staticmethod
    def list_data_sources(config=None):
        """This interface does not list data sources."""
        return []

    @staticmethod
    def query(uri, pattern, session_id=None, config=None):
        """Query a STIX bundle locally or remotely."""
        scheme, _, data_paths = uri.rpartition("://")
        data_paths = data_paths.split(",")
        pattern = fixup_pattern(pattern)
        compiled_pattern = Pattern(pattern)
        query_id = _make_query_id(uri, pattern)
        downloaddir = _make_download_dir()

        try:
            ingestdir = _make_query_dir(query_id)
        except FileExistsError:
            # We already cached this bundle
            data_paths = []

        bundles = []
        for i, data_path in enumerate(data_paths):
            data_path_stripped = "".join(filter(str.isalnum, data_path))
            ingestfile = ingestdir / f"{i}.json"

            if scheme == "file":
                bundlefile = data_path
                try:
                    with open(data_path, "r") as f:
                        bundle_in = json.load(f)
                except Exception:
                    raise DataSourceConnectionError(uri)
            elif scheme == "http" or scheme == "https":
                bundlefile = downloaddir / f"{data_path_stripped}.json"
                data_uri = f"{scheme}://{data_path}"
                last_modified = None
                file_time = None
                if bundlefile.exists():
                    _logger.debug("File exists: %s", bundlefile)
                    try:
                        resp = requests.head(data_uri)
                    except requests.exceptions.ConnectionError:
                        raise DataSourceConnectionError(uri)
                    last_modified = resp.headers.get("Last-Modified")
                    if last_modified:
                        last_modified = parser.parse(last_modified)
                        file_time = datetime.fromtimestamp(
                            bundlefile.stat().st_mtime, tz=timezone.utc
                        )
                    else:
                        _logger.debug(
                            "HTTP/HTTPS response header does not have 'Last-Modified' field"
                        )
                else:
                    _logger.debug("Bundle not on disk: %s", bundlefile)

                if not last_modified or last_modified > file_time:
                    _logger.info("Downloading %s to %s", data_uri, bundlefile)
                    try:
                        bundle_in = requests.get(data_uri).json()
                    except requests.exceptions.ConnectionError:
                        raise DataSourceConnectionError(uri)
                    with bundlefile.open("w") as f:
                        json.dump(bundle_in, f)
                else:
                    # We already have this file
                    _logger.debug("Using cached bundle: %s", bundlefile)
                    try:
                        with open(bundlefile, "r") as f:
                            bundle_in = json.load(f)
                    except Exception:
                        raise DataSourceConnectionError(uri)
            else:
                raise DataSourceManagerInternalError(
                    f"interface {__package__} should not process scheme {scheme}"
                )

            bundle_out = {}
            _logger.debug("Filtering: %s", bundlefile)
            count = 0
            matched = 0
            for prop, val in bundle_in.items():
                if prop == "objects":
                    bundle_out[prop] = []
                    for obj in val:
                        count += 1
                        if obj["type"] != "observed-data" or compiled_pattern.match(
                            [obj], False
                        ):
                            bundle_out[prop].append(obj)
                            matched += 1
                else:
                    bundle_out[prop] = val
            _logger.debug(
                "Matched %d of %d observations: %s", matched, count, bundlefile
            )

            with ingestfile.open("w") as f:
                json.dump(bundle_out, f)
            bundles.append(str(ingestfile.expanduser().resolve()))

        return ReturnFromFile(query_id, bundles)



"""The abstract interface for building a data source interface for Kestrel.

A Kestrel data source interface is a Python package with the following rules:

- The package name should use prefix ``kestrel_datasource_``.

- The package should have one and only one root level class inherited from
  :class:`AbstractDataSourceInterface`.

  - There is no restriction on package structure for the package.

  - There is no restriction on interface class name.

  - The interface class should inhert :class:`AbstractDataSourceInterface`.

  - The interface class should be importable from the package directly, i.e.,
    it needs to be imported into ``__init__.py`` of the package.

  - Zero class inherited from :class:`AbstractDataSourceInterface` will
    result in an exception.

  - Multiple classes inherited from :class:`AbstractDataSourceInterface` will
    result in an exception.

"""

from abc import ABC, abstractmethod

MODULE_PREFIX = "kestrel_datasource_"


class AbstractDataSourceInterface(ABC):
    """The abstract class for building a data source interface.

    Why do we design the interface this way? Actually we do not need a class
    for building the interface since all methods are static. However, in
    Python, we need to have a class if we'd like to enforce developers to
    implement the methods when developing a concrete interface. This is done by
    using both ``@staticmethod`` and ``@abstractmethod`` decorators for all
    methods/functions. When using an interface, Kestrel runtime will not
    instantiate an object from an interface class but use the static methods
    directly. This may not look beautiful in design, and hope we have
    something comparable to ``typeclass`` in Haskell for non-OOP interface
    abstraction in the future.

    """

    @staticmethod
    @abstractmethod
    def schemes():
        """``scheme`` (the URI prefix before ``://``) of the data source
        interface.

        Every data source interface should have at least one *unique* scheme to
        use at the beginning of the data source URI. To develop a new data
        source, one needs to check public Kestrel data source packages to name a new
        one that is not taken. Note that scheme defined here should be in
        lowercase, and Kestrel data source manager will normalize schemes of
        incoming URIs into lowercase.

        Returns:
            [str]: A list of schemes; A URI with one of the scheme will be
            processed by this interface.

        """
        return []

    @staticmethod
    @abstractmethod
    def list_data_sources(config):
        """List data source names accessible from this interface.

        Args:
            config (dict): a layered list/dict that contains config for the
              interface and can be edited/updated by the interface.

        Returns:
            [str]: A list of data source names accessible from this interface.

        """
        return []

    @staticmethod
    @abstractmethod
    def query(uri, pattern, session_id, config):
        """Sending a data query to a specific data source.

        If the store of the session is modified and directly gets the data
        loaded into a ``query_id``, it should return
        :attr:`kestrel.datasource.ReturnFromStore`.

        If the interface uses local files as intermediate/temporary storage
        before loading it to the store, it should return
        :attr:`kestrel.datasource.ReturnFromFile`.

        Args:
            uri (str): the full URI including the scheme and data source name.

            pattern (str): the pattern to query (currently we support STIX).

            session_id (str): id of the session, may be useful for analytics
              directly writing into the store.

            config (dict): a layered list/dict that contains config for the
              interface and can be edited/updated by the interface.

        Returns:
            kestrel.datasource.retstruct.AbstractReturnStruct: returned data.
            Currently there are two choices:
            :attr:`kestrel.datasource.ReturnFromFile` and
            :attr:`kestrel.datasource.ReturnFromStore`.

        """
        return None



from lark import Lark
from pkgutil import get_data
from itertools import chain

from kestrel.codegen.relations import (
    all_relations,
    stix_2_0_ref_mapping,
    stix_2_0_identical_mapping,
)


LITERALS = {"CNAME", "LETTER", "DIGIT", "WS", "INT", "WORD", "ESCAPED_STRING", "NUMBER"}
AGG_FUNCS = {"MIN", "MAX", "AVG", "SUM", "COUNT", "NUNIQUE"}
TRANSFORMS = {"TIMESTAMPED"}


def get_keywords():
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    parser = Lark(grammar, parser="lalr")
    alphabet_patterns = filter(lambda x: x.pattern.value.isalnum(), parser.terminals)
    keywords = [x.pattern.value for x in alphabet_patterns] + all_relations
    keywords_lower = map(lambda x: x.lower(), keywords)
    keywords_upper = map(lambda x: x.upper(), keywords)
    keywords_comprehensive = list(chain(keywords_lower, keywords_upper))
    return keywords_comprehensive


def get_entity_types():
    all_types = {"x-ibm-finding", "x-oca-asset", "x-oca-event"}
    for mapping in stix_2_0_ref_mapping:
        for i in (0, 2):
            if not mapping[i].endswith("-ext"):
                all_types.add(mapping[i])
    all_types.update(stix_2_0_identical_mapping.keys())
    return tuple(all_types)



import ast
from datetime import datetime, timedelta
from pkgutil import get_data

from firepit.query import BinnedColumn, Filter, Predicate
from firepit.timestamp import timefmt
from lark import Lark, Token, Transformer, Tree


def parse(stmts, default_variable="_", default_sort_order="desc"):
    # the public parsing interface for Kestrel
    # return abstract syntax tree
    # check kestrel.lark for details
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    return Lark(
        grammar,
        parser="lalr",
        transformer=_PostParsing(default_variable, default_sort_order),
    ).parse(stmts)


def get_all_input_var_names(stmt):
    input_refs = ["input", "input_2", "variablesource"]
    inputs_refs = stmt["inputs"] if "inputs" in stmt else []
    return [stmt.get(k) for k in input_refs if k in stmt] + inputs_refs


################################################################
#                           Private
################################################################


class _PostParsing(Transformer):
    def __init__(self, default_variable, default_sort_order):
        self.default_variable = default_variable
        self.default_sort_order = default_sort_order
        super().__init__()

    def start(self, args):
        return args

    def statement(self, args):
        # Kestrel syntax: a statement can only has one command
        stmt = args.pop()
        return stmt

    def assignment(self, args):
        stmt = args[1] if len(args) == 2 else args[0]
        stmt["output"] = _extract_var(args, self.default_variable)
        return stmt

    def assign(self, args):
        result = args[0]  # Already transformed in expression method below
        result["command"] = "assign"
        return result

    def merge(self, args):
        return {
            "command": "merge",
            "inputs": _extract_vars(args, self.default_variable),
        }

    def info(self, args):
        return {"command": "info", "input": _extract_var(args, self.default_variable)}

    def disp(self, args):
        result = {"command": "disp"}
        for arg in args:
            if isinstance(arg, dict):
                result.update(arg)
        if "attrs" not in result:
            result["attrs"] = "*"
        return result

    def get(self, args):
        datasource = _extract_datasource(args)
        packet = {
            "command": "get",
            "type": _extract_entity_type(args),
            "patternbody": _assert_and_extract_single("STIXPATTERNBODY", args),
        }
        if datasource:
            packet["datasource"] = datasource

        time_range = {}
        for item in args:
            if isinstance(item, dict):
                time_range.update(item)

        if "start" in time_range and "stop" in time_range:
            packet["timerange"] = (time_range["start"], time_range["stop"])
        else:
            packet["timerange"] = None

        return packet

    def find(self, args):
        packet = {
            "command": "find",
            "type": _extract_entity_type(args),
            "relation": _assert_and_extract_single("RELATION", args).lower(),
            "reversed": _extract_if_reversed(args),
            "input": _extract_var(args, self.default_variable),
        }

        time_range = {}
        for item in args:
            if isinstance(item, dict):
                time_range.update(item)

        if "start" in time_range and "stop" in time_range:
            packet["timerange"] = (time_range["start"], time_range["stop"])
        else:
            packet["timerange"] = None

        return packet

    def join(self, args):
        if len(args) == 5:
            return {
                "command": "join",
                "input": _first(args),
                "input_2": _second(args),
                "path": _fourth(args),
                "path_2": _fifth(args),
            }
        else:
            return {"command": "join", "input": _first(args), "input_2": _second(args)}

    def group(self, args):
        # args[1] was already transformed by path_list/valuelist
        cols = args[2]
        result = {
            "command": "group",
            "paths": cols,
            "input": _extract_var(args, self.default_variable),
        }
        aggregations = args[3] if len(args) > 3 else None
        if aggregations:
            result["aggregations"] = aggregations
        return result

    def sort(self, args):
        return {
            "command": "sort",
            "path": _extract_stixpath(args),
            "input": _extract_var(args, self.default_variable),
            "ascending": _extract_direction(args, self.default_sort_order),
        }

    def apply(self, args):
        input_vars = []
        for arg in args:
            if isinstance(arg, dict) and "variables" in arg:
                input_vars = arg["variables"]
                break
        return {
            "command": "apply",
            "workflow": _first(args),
            "inputs": input_vars,
            "parameter": _merge_args("anaparams", args),
        }

    def load(self, args):
        return {
            "command": "load",
            "type": _extract_entity_type(args),
            "path": _extract_dumppath(args),
        }

    def save(self, args):
        return {
            "command": "save",
            "input": _extract_var(args, self.default_variable),
            "path": _extract_dumppath(args),
        }

    def new(self, args):
        return {
            "command": "new",
            "type": _extract_entity_type(args),
            "data": _assert_and_extract_single("VARDATA", args),
        }

    def expression(self, args):
        result = args[0]
        for arg in args:
            result.update(arg)
        return result

    def transform(self, args):
        return {
            "input": _extract_var(args, self.default_variable),
            "transform": _assert_and_extract_single("TRANSFORM", args),
        }

    def where_clause(self, args):
        return {
            "where": Filter([args[0]]),
        }

    def attr_clause(self, args):
        paths = _assert_and_extract_single("STIXPATHS", args)
        return {
            "attrs": paths if paths else "*",
        }

    def sort_clause(self, args):
        return {
            "path": _extract_stixpath(args),
            "ascending": _extract_direction(args, self.default_sort_order),
        }

    def limit_clause(self, args):
        return {
            "limit": _extract_int(args),
        }

    def offset_clause(self, args):
        return {
            "offset": _extract_int(args),
        }

    def timespan(self, args):
        result = {}
        for arg in args:
            result.update(arg)
        return result

    def relative_timespan(self, args):
        # args should be num and unit
        num = int(_first(args))
        unit = _second(args).lower()
        if unit in ("days", "d"):
            delta = timedelta(days=num)
        elif unit in ("hours", "h"):
            delta = timedelta(hours=num)
        elif unit in ("minutes", "m"):
            delta = timedelta(minutes=num)
        elif unit in ("seconds", "s"):
            delta = timedelta(seconds=num)
        stop = datetime.utcnow()
        start = stop - delta
        return {"start": timefmt(start, prec=6), "stop": timefmt(stop, prec=6)}

    def starttime(self, args):
        return {"start": _assert_and_extract_single("ISOTIMESTAMP", args)}

    def endtime(self, args):
        return {"stop": _assert_and_extract_single("ISOTIMESTAMP", args)}

    def variables(self, args):
        return {"variables": _extract_vars(args, self.default_variable)}

    def localargs(self, args):
        return {args[0].value: args[1]}

    def valuelist(self, args):
        if len(args) == 1:
            return args[0].value
        elif len(args) > 1:
            return [arg.value for arg in args]
        return None

    def number(self, args):
        return {args[0].value: ast.literal_eval(args[1].value)}

    def grp_spec(self, args):
        return args

    def grp_expr(self, args):
        path = args[0]
        if isinstance(path, str):
            path = _normalize_paths(path)[0]
        return path

    def bin_func(self, args):
        path = args[0].value
        num = int(args[1].value)
        if len(args) >= 3:
            unit = args[2].value
        else:
            unit = None
        alias = f"{path}_bin"
        return BinnedColumn(path, num, unit, alias=alias)

    def agg_list(self, args):
        return [arg for arg in args]

    def agg(self, args):
        func = args[0].value.lower()
        alias = args[2].value if len(args) > 2 else f"{func}_{args[1].value}"
        return {"func": func, "attr": args[1].value, "alias": alias}

    def disj(self, args):
        lhs = str(args[0]) if isinstance(args, Token) else args[0]
        rhs = str(args[1]) if isinstance(args, Token) else args[1]
        return Predicate(lhs, "OR", rhs)

    def conj(self, args):
        lhs = str(args[0]) if isinstance(args, Token) else args[0]
        rhs = str(args[1]) if isinstance(args, Token) else args[1]
        return Predicate(lhs, "AND", rhs)

    def comp(self, args):
        lhs = str(args[0]) if isinstance(args, Token) else args[0]
        op = str(args[1])
        rhs = str(args[2]) if isinstance(args, Token) else args[2]
        return Predicate(lhs, op, rhs)

    def null_comp(self, args):
        lhs = str(args[0])
        op = str(args[1])
        if op == "IS NOT":
            op = "!="
        else:
            op = "="
        rhs = "NULL"
        return Predicate(lhs, op, rhs)

    def column(self, args):
        return args[0].value

    def squoted_str(self, args):
        return args[0].value.strip("'")

    def num_literal(self, args):
        return args[0].value

    def null(self, args):
        return "NULL"


def _first(args):
    return args[0].value


def _second(args):
    return args[1].value


def _third(args):
    return args[2].value


def _fourth(args):
    return args[3].value


def _fifth(args):
    return args[4].value


def _last(args):
    return args[-1].value


def _assert_and_extract_single(arg_type, args):
    items = [arg.value for arg in args if hasattr(arg, "type") and arg.type == arg_type]
    assert len(items) <= 1
    return items.pop() if items else None


def _extract_int(args):
    value = _assert_and_extract_single("INT", args)
    return int(value) if value else None


def _extract_var(args, default_variable):
    # extract a single variable from the args
    # default variable if no variable is found
    v = _assert_and_extract_single("VARIABLE", args)
    return v if v else default_variable


def _extract_vars(args, default_variable):
    var_names = []
    for arg in args:
        if hasattr(arg, "type") and arg.type == "VARIABLE":
            var_names.append(arg.value)
    if not var_names:
        var_names = [default_variable]
    return var_names


def _extract_stixpath(args):
    # extract a single stix path from the args
    return _assert_and_extract_single("STIXPATH", args)


def _extract_datasource(args):
    raw_ds = _assert_and_extract_single("DATASRC", args)
    ds = raw_ds.strip('"') if raw_ds else None
    return ds


def _extract_entity_type(args):
    # extract a single entity type from the args
    return _assert_and_extract_single("ENTITY_TYPE", args)


def _extract_dumppath(args):
    # extract a single stix path from the args
    return _assert_and_extract_single("DUMPPATH", args)


def _extract_direction(args, default_sort_order):
    # extract sort direction from args
    # default direction if no variable is found
    # return: if descending
    ds = [
        x for x in args if hasattr(x, "type") and (x.type == "ASC" or x.type == "DESC")
    ]
    assert len(ds) <= 1
    d = ds.pop().type if ds else default_sort_order
    return True if d == "ASC" else False


def _extract_if_reversed(args):
    rs = [x for x in args if hasattr(x, "type") and x.type == "REVERSED"]
    return True if rs else False


def _merge_args(arg_type, args):
    items = dict()
    for arg in args:
        if not isinstance(arg, Tree) or not hasattr(arg, "data"):
            continue
        if arg.data != arg_type:
            continue
        for t in arg.children:
            items.update(t)
    return items


def _normalize_paths(stixpaths):
    if isinstance(stixpaths, str):
        stixpaths = [stixpaths]
    # Sanitize paths
    cols = []
    for stixpath in stixpaths:
        sco_type, _, prop = stixpath.rpartition(":")
        # We should probably verify sco_type here
        cols.append(prop)
    return cols



import dateutil.parser
import datetime
import logging
import re

from kestrel.semantics import get_entity_table
from kestrel.syntax.paramstix import parse_extended_stix_pattern
from kestrel.exceptions import (
    InvalidAttribute,
    UnsupportedStixSyntax,
    KestrelInternalError,
)
from firepit.exceptions import InvalidAttr

_logger = logging.getLogger(__name__)


def or_patterns(patterns):
    bodies = []
    time_range = []
    for pattern in patterns:
        if pattern:
            pieces = pattern.split()
            if len(pieces) > 4 and pieces[-4] == "START" and pieces[-2] == "STOP":
                time_range.append((pieces[-3], pieces[-1]))
                bodies.append("(" + " ".join(pieces[:-4]) + ")")
            else:
                bodies.append(pattern)
    if bodies:
        if time_range:
            start = min([t[0] for t in time_range])
            end = max([t[1] for t in time_range])
            final_pattern = (
                "(" + " OR ".join(bodies) + ")" + " START " + start + " STOP " + end
            )
        else:
            final_pattern = " OR ".join(bodies)
        _logger.debug(f"or pattern merged: {final_pattern}")
    else:
        final_pattern = None
        _logger.info(f"all None patterns input into or_patterns()")

    return final_pattern


def build_pattern(
    raw_pattern_body, time_range, start_offset, end_offset, symtable, store
):
    """Dereference variables in a STIX pattern and output the unfolded pattern."""
    references = parse_extended_stix_pattern(raw_pattern_body)

    pattern_body = raw_pattern_body

    _logger.debug(f"building pattern for: {raw_pattern_body}")

    if references:
        _logger.debug(f"references found: {list(references.keys())}")

        var_attr_to_vals_str = _dereference_multiple_variables(
            store, symtable, references
        )
        for var_attr, vals_str in var_attr_to_vals_str.items():
            pattern_body = _replace_ref_with_op(pattern_body, var_attr, vals_str)

        _logger.debug(f'pattern body dereferred: "{pattern_body}"')

        if pattern_body and not time_range:
            ref_var_time_ranges = []

            for var_name in references.keys():
                try:
                    _tr = _get_variable_time_range(store, symtable, var_name)
                except InvalidAttribute:
                    _tr = None
                    _logger.warning(
                        f"pattern time range searching failed on variable {var_name}"
                    )
                else:
                    ref_var_time_ranges.append(_tr)

            if ref_var_time_ranges:
                start = min([t[0] for t in ref_var_time_ranges])
                end = max([t[1] for t in ref_var_time_ranges])

                start_adj = start + datetime.timedelta(seconds=start_offset)
                end_adj = end + datetime.timedelta(seconds=end_offset)

                start_stix = start_adj.strftime("%Y-%m-%dT%H:%M:%S.000Z")
                stop_stix = end_adj.strftime("%Y-%m-%dT%H:%M:%S.000Z")

                time_range = (start_stix, stop_stix)
                _logger.debug(f"pattern time range computed: {time_range}")

    if pattern_body:
        if time_range:
            pattern = (
                f"({pattern_body}) START t'{time_range[0]}' STOP t'{time_range[1]}'"
            )
        else:
            pattern = f"{pattern_body}"
        _logger.debug(f'final pattern assembled: "{pattern}"')
    else:
        pattern = None
        _logger.info(f"empty pattern assembled")

    return pattern


def build_pattern_from_ids(return_type, ids):
    if ids:
        return "[" + return_type + ":id IN (" + ", ".join(map(_type_value, ids)) + ")]"
    else:
        return None


def _dereference_multiple_variables(store, symtable, references):
    return {
        var + "." + attr: "(" + ", ".join(map(_type_value, vals)) + ")"
        for var, attrs in references.items()
        for attr, vals in _dereference_variable(store, symtable, var, attrs).items()
    }


def _dereference_variable(store, symtable, var_name, attributes):
    attr_line = ",".join(attributes)
    _logger.debug(f'deref "{var_name}" with attributes "{attr_line}"')

    var_entity_table = get_entity_table(var_name, symtable)
    try:
        store_return = store.lookup(var_entity_table, attr_line)
    except InvalidAttr as e:
        _logger.warning(f"cannot deref {attr_line}. Invalid attribute in firepit.")
        raise InvalidAttribute(e.message)

    attr_to_values = {k: [] for k in attributes}
    for row in store_return:
        for k, v in row.items():
            if v is not None and v not in attr_to_values[k]:
                attr_to_values[k].append(v)

    for k, v in attr_to_values.items():
        if not v:
            raise InvalidAttribute(var_name + "." + k)

    _logger.debug(f"deref results: {str(attr_to_values)}")

    return attr_to_values


def _get_variable_time_range(store, symtable, var_name):
    """
    Returns:
        start (datetime.datetime): the time any entities is observed first.
        end (datetime.datetime): the time any entities is observed last.

    """
    var_entity_table = get_entity_table(var_name, symtable)
    summary = store.summary(var_entity_table)
    start = dateutil.parser.isoparse(summary["first_observed"])
    end = dateutil.parser.isoparse(summary["last_observed"])
    return start, end


def _type_value(value):
    if isinstance(value, str):
        return f"'{value}'"
    elif isinstance(value, int):
        return str(value)
    elif isinstance(value, float):
        # pandas dataframe and sqlite may save integers as floats
        return str(round(value))
    else:
        return str(value)


def _replace_ref_with_op(pattern, var_attr, vals_str):

    # avoid adhesive parans/ops that prevent correct splitting
    pattern = re.sub(r"([=><\[\]])", r" \1 ", pattern)

    pieces = pattern.split()
    try:
        ref_index = pieces.index(var_attr)
    except ValueError:
        err_msg = f'cannot find "{var_attr}" when assembling pattern "{pattern}"'
        _logger.error(err_msg)
        raise KestrelInternalError(err_msg)
    if pieces[ref_index - 1] == "=":
        pieces[ref_index - 1] = "IN"
        pieces[ref_index] = vals_str
    else:
        raise UnsupportedStixSyntax(
            'only "=" is supported before referred variable in parameterized STIX'
        )
    return " ".join(pieces)
